@@ -3,11 +3,10 @@
 # Copyright (c) 2019 Red Hat, Inc.
 # Copyright (c) 2019 Tomáš Mráz <tmraz@fedoraproject.org>
 
-from subprocess import check_output, CalledProcessError
+from subprocess import CalledProcessError, check_output
 
 from .configgenerator import ConfigGenerator
 
-
 RH_ALLOW_SHA1 = '''
 [openssl_init]
 alg_section = evp_properties
@@ -45,6 +44,16 @@ class OpenSSLGenerator(ConfigGenerator):
         'RC2-CBC': '!RC2',
         'NULL': '!eNULL:!aNULL'
     }
+    cipher_notany_multimap = {
+        # CBC is documented as SSL_DES, SSL_3DES, SSL_RC2, SSL_IDEA,
+        #                      SSL_AES128, SSL_AES256,
+        #                      SSL_CAMELLIA128, SSL_CAMELLIA256, SSL_SEED
+        '-CBC': {'DES-CBC', '3DES-CBC', 'RC2-CBC', 'IDEA-CBC',
+                 'AES-128-CBC', 'AES-256-CBC',
+                 'CAMELLIA-128-CBC', 'CAMELLIA-256-CBC', 'SEED-CBC'},
+        '-AESCCM': {'AES-128-CCM', 'AES-256-CCM'},
+        '-AESGCM': {'AES-128-GCM', 'AES-256-GCM'},
+    }
 
     key_exchange_map = {
         'RSA': 'kRSA',
@@ -78,13 +87,44 @@ class OpenSSLGenerator(ConfigGenerator):
     }
 
     ciphersuite_map = {
-        'AES-256-GCM': 'TLS_AES_256_GCM_SHA384',
-        'AES-128-GCM': 'TLS_AES_128_GCM_SHA256',
-        'CHACHA20-POLY1305': 'TLS_CHACHA20_POLY1305_SHA256',
-        'AES-128-CCM': 'TLS_AES_128_CCM_SHA256',
-        'AES-128-CCM8': 'TLS_AES_128_CCM_8_SHA256',
-        'GOST28147-TC26Z-CNT': 'GOST2012-GOST8912-GOST8912',
-        'GOST28147-CPA-CNT': 'GOST2001-GOST89-GOST89'
+        'TLS_AES_256_GCM_SHA384': {
+            'cipher': {'AES-256-GCM'},
+            'hash': {'SHA2-384'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_AES_128_GCM_SHA256': {
+            'cipher': {'AES-128-GCM'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_CHACHA20_POLY1305_SHA256': {
+            'cipher': {'CHACHA20-POLY1305'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_AES_128_CCM_SHA256': {
+            'cipher': {'AES-128-CCM'},
+            'hash': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        # This one is not enableable since c-p does not expose CCM8 ciphers:
+        # 'TLS_AES_128_CCM_8_SHA256': {
+        #     'cipher': {'AES-128-CCM8'},  # this is not a thing in c-p
+        #     'hash': {'SHA2-256'},
+        #     'protocol': {'TLS1.3', 'DTLS1.3'},
+        # },
+        'TLS_SHA256_SHA256': {
+            'cipher': {'NULL'},
+            'hash': {'SHA2-256'},
+            'mac': {'SHA2-256'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
+        'TLS_SHA256_SHA384': {
+            'cipher': {'NULL'},
+            'hash': {'SHA2-384'},
+            'mac': {'SHA2-384'},
+            'protocol': {'TLS1.3', 'DTLS1.3'},
+        },
     }
 
     @classmethod
@@ -121,9 +161,9 @@ class OpenSSLGenerator(ConfigGenerator):
                 s = cls.append(s, cls.cipher_not_map[i])
             except KeyError:
                 pass
-        if 'AES-128-CCM' in ip['cipher']:
-            if 'AES-256-CCM' in ip['cipher']:
-                s = cls.append(s, '-AESCCM')
+        for keyword, cipherset in cls.cipher_notany_multimap.items():
+            if all(c in ip['cipher'] for c in cipherset):
+                s = cls.append(s, keyword)
 
         for i in ip['mac']:
             try:
@@ -134,10 +174,8 @@ class OpenSSLGenerator(ConfigGenerator):
         # These ciphers are not necessary for any
         # policy level, and only increase the attack surface.
         # FIXME! must be fixed for custom policies
-        s = cls.append(s, '-SHA384')
-        s = cls.append(s, '-CAMELLIA')
-        s = cls.append(s, '-ARIA')
-        s = cls.append(s, '-AESCCM8')
+        for c in ('-SHA384', '-CAMELLIA', '-ARIA', '-AESCCM8'):
+            s = cls.append(s, c)
 
         return s
 
@@ -145,12 +183,16 @@ class OpenSSLGenerator(ConfigGenerator):
     def generate_ciphersuites(cls, policy):
         s = ''
         p = policy.enabled
-        for i in p['cipher']:
-            try:
-                s = cls.append(s, cls.ciphersuite_map[i])
-            except KeyError:
-                pass
-
+        # we need the output sorted by order of `cipher`
+        # finer sorting nuances are currently ignored
+        for c in p['cipher']:
+            cipher_submap = {name: spec for name, spec
+                             in cls.ciphersuite_map.items()
+                             if spec['cipher'] == {c}}
+            for ciphersuite_name, ciphersuite_spec in cipher_submap.items():
+                if all(any(val in algvalues for val in p[algclass])
+                       for algclass, algvalues in ciphersuite_spec.items()):
+                    s = cls.append(s, ciphersuite_name)
         return s
 
     @classmethod
@@ -160,9 +202,10 @@ class OpenSSLGenerator(ConfigGenerator):
     @classmethod
     def test_config(cls, config):
         output = b''
-        assert config.endswith('\n')
+        assert config.endswith('\n')  # noqa: S101
         try:
-            output = check_output(['openssl', 'ciphers', config[:-1]])
+            output = check_output(['openssl',  # noqa: S607
+                                   'ciphers', config[:-1]])
         except CalledProcessError:
             cls.eprint('There is an error in openssl generated policy')
             cls.eprint(f'Policy:\n{config}')
